@@ -24,6 +24,8 @@
 #include "init/progress.h"
 #include "init/recovery.h"
 
+#include "sb/counters.h"
+
 #include "util/clock.h"
 #include "util/enumerated_ref.h"
 #include "util/varint.h"
@@ -348,7 +350,6 @@ static inline void __bch2_alloc_v4_to_text(struct printbuf *out, struct bch_fs *
 	struct bch_dev *ca = c ? bch2_dev_tryget_noerror(c, k.k->p.inode) : NULL;
 
 	prt_newline(out);
-	guard(printbuf_indent)(out);
 
 	prt_printf(out, "gen %u oldest_gen %u data_type ", a->gen, a->oldest_gen);
 	bch2_prt_data_type(out, a->data_type);
@@ -922,7 +923,7 @@ int bch2_trigger_alloc(struct btree_trans *trans,
 
 		if (statechange(a->data_type == BCH_DATA_free) &&
 		    bucket_flushed(new_a))
-			closure_wake_up(&c->freelist_wait);
+			closure_wake_up(&c->allocator.freelist_wait);
 
 		if (statechange(a->data_type == BCH_DATA_need_discard) &&
 		    !bch2_bucket_is_open_safe(c, new.k->p.inode, new.k->p.offset) &&
@@ -976,6 +977,28 @@ static void discard_in_flight_remove(struct bch_dev *ca, u64 bucket)
 	darray_remove_item(&ca->discard_buckets_in_flight, i);
 }
 
+struct discard_buckets_state {
+	u64		seen;
+	u64		open;
+	u64		need_journal_commit;
+	u64		commit_in_flight;
+	u64		bad_data_type;
+	u64		already_discarding;
+	u64		discarded;
+};
+
+static void discard_buckets_state_to_text(struct printbuf *out, struct discard_buckets_state *s)
+{
+	printbuf_tabstop_push(out, 20);
+	prt_printf(out, "seen:\t%llu\n",		s->seen);
+	prt_printf(out, "open:\t%llu\n",		s->open);
+	prt_printf(out, "need_journal_commit:\t%llu\n",	s->need_journal_commit);
+	prt_printf(out, "commit_in_flight:\t%llu\n",	s->commit_in_flight);
+	prt_printf(out, "bad_data_type:\t%llu\n",	s->bad_data_type);
+	prt_printf(out, "already_discarding:\t%llu\n",	s->already_discarding);
+	prt_printf(out, "discarded:\t%llu\n",		s->discarded);
+}
+
 static int bch2_discard_one_bucket(struct btree_trans *trans,
 				   struct bch_dev *ca,
 				   struct btree_iter *need_discard_iter,
@@ -1008,6 +1031,10 @@ static int bch2_discard_one_bucket(struct btree_trans *trans,
 	CLASS(btree_iter, iter)(trans, BTREE_ID_alloc, need_discard_iter->pos, BTREE_ITER_cached);
 	struct bkey_s_c k = bkey_try(bch2_btree_iter_peek_slot(&iter));
 
+	struct bkey_buf orig_k __cleanup(bch2_bkey_buf_exit);
+	bch2_bkey_buf_init(&orig_k);
+	bch2_bkey_buf_reassemble(&orig_k, k);
+
 	struct bkey_i_alloc_v4 *a = errptr_try(bch2_alloc_to_v4_mut(trans, k));
 
 	if (a->v.data_type != BCH_DATA_need_discard) {
@@ -1065,9 +1092,11 @@ commit:
 		goto out;
 
 	if (!fastpath)
-		count_event(c, bucket_discard);
+		event_inc_trace(c, bucket_discard, buf,
+			bch2_bkey_val_to_text(&buf, c, bkey_i_to_s_c(orig_k.k)));
 	else
-		count_event(c, bucket_discard_fast);
+		event_inc_trace(c, bucket_discard_fast, buf,
+			bch2_bkey_val_to_text(&buf, c, bkey_i_to_s_c(orig_k.k)));
 out:
 fsck_err:
 	if (discard_locked)
@@ -1097,7 +1126,10 @@ static void __bch2_dev_do_discards(struct bch_dev *ca)
 	if (s.need_journal_commit > dev_buckets_available(ca, BCH_WATERMARK_normal))
 		bch2_journal_flush_async(&c->journal, NULL);
 
-	trace_discard_buckets(c, &s, bch2_err_str(ret));
+	event_inc_trace(c, bucket_discard_worker, buf, ({
+		prt_printf(&buf, "ret %s\ndev %s\n", bch2_err_str(ret), ca->name);
+		discard_buckets_state_to_text(&buf, &s);
+	}));
 
 	enumerated_ref_put(&ca->io_ref[WRITE], BCH_DEV_WRITE_REF_dev_do_discards);
 }
@@ -1201,7 +1233,10 @@ static void bch2_do_discards_fast_work(struct work_struct *work)
 			break;
 	}
 
-	trace_discard_buckets_fast(c, &s, bch2_err_str(ret));
+	event_inc_trace(c, bucket_discard_fast_worker, buf, ({
+		prt_printf(&buf, "ret %s\ndev %s\n", bch2_err_str(ret), ca->name);
+		discard_buckets_state_to_text(&buf, &s);
+	}));
 
 	bch2_trans_put(trans);
 	enumerated_ref_put(&ca->io_ref[WRITE], BCH_DEV_WRITE_REF_discard_one_bucket_fast);
@@ -1283,7 +1318,6 @@ static int invalidate_one_bucket(struct btree_trans *trans,
 				 s64 *nr_to_invalidate)
 {
 	struct bch_fs *c = trans->c;
-	CLASS(printbuf, buf)();
 	struct bpos bucket = u64_to_bucket(lru_k.k->p.offset);
 
 	if (*nr_to_invalidate <= 0)
@@ -1300,42 +1334,44 @@ static int invalidate_one_bucket(struct btree_trans *trans,
 	if (bch2_bucket_is_open_safe(c, bucket.inode, bucket.offset))
 		return 0;
 
-	{
-		CLASS(btree_iter, alloc_iter)(trans, BTREE_ID_alloc, bucket, BTREE_ITER_cached);
-		struct bkey_s_c alloc_k = bkey_try(bch2_btree_iter_peek_slot(&alloc_iter));
+	CLASS(btree_iter, alloc_iter)(trans, BTREE_ID_alloc, bucket, BTREE_ITER_cached);
+	struct bkey_s_c alloc_k = bkey_try(bch2_btree_iter_peek_slot(&alloc_iter));
 
-		struct bch_alloc_v4 a_convert;
-		const struct bch_alloc_v4 *a = bch2_alloc_to_v4(alloc_k, &a_convert);
+	struct bch_alloc_v4 a_convert;
+	const struct bch_alloc_v4 *a = bch2_alloc_to_v4(alloc_k, &a_convert);
 
-		/* We expect harmless races here due to the btree write buffer: */
-		if (lru_pos_time(lru_iter->pos) != alloc_lru_idx_read(*a))
-			return 0;
+	/* We expect harmless races here due to the btree write buffer: */
+	if (lru_pos_time(lru_iter->pos) != alloc_lru_idx_read(*a))
+		return 0;
 
-		/*
-		 * Impossible since alloc_lru_idx_read() only returns nonzero if the
-		 * bucket is supposed to be on the cached bucket LRU (i.e.
-		 * BCH_DATA_cached)
-		 *
-		 * bch2_lru_validate() also disallows lru keys with lru_pos_time() == 0
-		 */
-		BUG_ON(a->data_type != BCH_DATA_cached);
-		BUG_ON(a->dirty_sectors);
+	/*
+	 * Impossible since alloc_lru_idx_read() only returns nonzero if the
+	 * bucket is supposed to be on the cached bucket LRU (i.e.
+	 * BCH_DATA_cached)
+	 *
+	 * bch2_lru_validate() also disallows lru keys with lru_pos_time() == 0
+	 */
+	BUG_ON(a->data_type != BCH_DATA_cached);
+	BUG_ON(a->dirty_sectors);
 
-		if (!a->cached_sectors) {
-			bch2_check_bucket_backpointer_mismatch(trans, ca, bucket.offset,
-							       true, last_flushed);
-			return 0;
-		}
-
-		unsigned cached_sectors = a->cached_sectors;
-		u8 gen = a->gen;
-
-		try(invalidate_one_bucket_by_bps(trans, ca, bucket, gen, last_flushed));
-
-		trace_and_count(c, bucket_invalidate, c, bucket.inode, bucket.offset, cached_sectors);
-		--*nr_to_invalidate;
+	if (!a->cached_sectors) {
+		bch2_check_bucket_backpointer_mismatch(trans, ca, bucket.offset,
+						       true, last_flushed);
+		return 0;
 	}
 
+	u8 gen = a->gen;
+
+	struct bkey_buf orig_alloc_k __cleanup(bch2_bkey_buf_exit);
+	bch2_bkey_buf_init(&orig_alloc_k);
+	bch2_bkey_buf_reassemble(&orig_alloc_k, alloc_k);
+
+	try(invalidate_one_bucket_by_bps(trans, ca, bucket, gen, last_flushed));
+
+	event_inc_trace(c, bucket_invalidate, buf,
+		bch2_bkey_val_to_text(&buf, c, bkey_i_to_s_c(orig_alloc_k.k)));
+
+	--*nr_to_invalidate;
 	return 0;
 }
 
@@ -1452,7 +1488,7 @@ int bch2_dev_remove_alloc(struct bch_fs *c, struct bch_dev *ca)
 		bch2_btree_delete_range(c, BTREE_ID_alloc, start, end,
 					BTREE_TRIGGER_norun) ?:
 		bch2_dev_usage_remove(c, ca);
-	bch_err_msg(ca, ret, "removing dev alloc info");
+	bch_err_msg_dev(ca, ret, "removing dev alloc info");
 	return ret;
 }
 
@@ -1551,13 +1587,13 @@ void bch2_recalc_capacity(struct bch_fs *c)
 
 	reserved_sectors = min(reserved_sectors, capacity);
 
-	c->reserved = reserved_sectors;
-	c->capacity = capacity - reserved_sectors;
+	c->capacity.reserved = reserved_sectors;
+	c->capacity.capacity = capacity - reserved_sectors;
 
-	c->bucket_size_max = bucket_size_max;
+	c->capacity.bucket_size_max = bucket_size_max;
 
 	/* Wake up case someone was waiting for buckets */
-	closure_wake_up(&c->freelist_wait);
+	closure_wake_up(&c->allocator.freelist_wait);
 }
 
 u64 bch2_min_rw_member_capacity(struct bch_fs *c)
@@ -1574,8 +1610,8 @@ static bool bch2_dev_has_open_write_point(struct bch_fs *c, struct bch_dev *ca)
 {
 	struct open_bucket *ob;
 
-	for (ob = c->open_buckets;
-	     ob < c->open_buckets + ARRAY_SIZE(c->open_buckets);
+	for (ob = c->allocator.open_buckets;
+	     ob < c->allocator.open_buckets + ARRAY_SIZE(c->allocator.open_buckets);
 	     ob++) {
 		scoped_guard(spinlock, &ob->lock) {
 			if (ob->valid && !ob->on_partial_list &&
@@ -1591,7 +1627,7 @@ void bch2_dev_allocator_set_rw(struct bch_fs *c, struct bch_dev *ca, bool rw)
 {
 	/* BCH_DATA_free == all rw devs */
 
-	for (unsigned i = 0; i < ARRAY_SIZE(c->rw_devs); i++) {
+	for (unsigned i = 0; i < ARRAY_SIZE(c->allocator.rw_devs); i++) {
 		bool data_type_rw = rw;
 
 		if (i != BCH_DATA_free &&
@@ -1603,10 +1639,10 @@ void bch2_dev_allocator_set_rw(struct bch_fs *c, struct bch_dev *ca, bool rw)
 		    !ca->mi.durability)
 			data_type_rw = false;
 
-		mod_bit(ca->dev_idx, c->rw_devs[i].d, data_type_rw);
+		mod_bit(ca->dev_idx, c->allocator.rw_devs[i].d, data_type_rw);
 	}
 
-	c->rw_devs_change_count++;
+	c->allocator.rw_devs_change_count++;
 }
 
 /* device goes ro: */
@@ -1628,7 +1664,7 @@ void bch2_dev_allocator_remove(struct bch_fs *c, struct bch_dev *ca)
 	 * Wake up threads that were blocked on allocation, so they can notice
 	 * the device can no longer be removed and the capacity has changed:
 	 */
-	closure_wake_up(&c->freelist_wait);
+	closure_wake_up(&c->allocator.freelist_wait);
 
 	/*
 	 * journal_res_get() can block waiting for free space in the journal -
@@ -1638,7 +1674,7 @@ void bch2_dev_allocator_remove(struct bch_fs *c, struct bch_dev *ca)
 
 	/* Now wait for any in flight writes: */
 
-	closure_wait_event(&c->open_buckets_wait,
+	closure_wait_event(&c->allocator.open_buckets_wait,
 			   !bch2_dev_has_open_write_point(c, ca));
 }
 
@@ -1648,7 +1684,7 @@ void bch2_dev_allocator_add(struct bch_fs *c, struct bch_dev *ca)
 	lockdep_assert_held(&c->state_lock);
 
 	bch2_dev_allocator_set_rw(c, ca, true);
-	c->rw_devs_change_count++;
+	c->allocator.rw_devs_change_count++;
 }
 
 void bch2_dev_allocator_background_exit(struct bch_dev *ca)
@@ -1666,5 +1702,31 @@ void bch2_dev_allocator_background_init(struct bch_dev *ca)
 
 void bch2_fs_allocator_background_init(struct bch_fs *c)
 {
-	spin_lock_init(&c->freelist_lock);
+	spin_lock_init(&c->allocator.freelist_lock);
+}
+
+void bch2_fs_capacity_exit(struct bch_fs *c)
+{
+	percpu_free_rwsem(&c->capacity.mark_lock);
+	if (c->capacity.pcpu) {
+		u64 v = percpu_u64_get(&c->capacity.pcpu->online_reserved);
+		WARN(v, "online_reserved not 0 at shutdown: %lli", v);
+	}
+
+	free_percpu(c->capacity.pcpu);
+	free_percpu(c->capacity.usage);
+}
+
+int bch2_fs_capacity_init(struct bch_fs *c)
+{
+	mutex_init(&c->capacity.sectors_available_lock);
+	seqcount_init(&c->capacity.usage_lock);
+
+	try(percpu_init_rwsem(&c->capacity.mark_lock));
+
+	if (!(c->capacity.pcpu = alloc_percpu(struct bch_fs_capacity_pcpu)) ||
+	    !(c->capacity.usage = alloc_percpu(struct bch_fs_usage_base)))
+		return bch_err_throw(c, ENOMEM_fs_other_alloc);
+
+	return 0;
 }
