@@ -57,7 +57,7 @@ static void bbuf_exit(struct bbuf *buf)
 		kfree(buf->b);
 		break;
 	case BB_mempool:
-		mempool_free(buf->b, &buf->c->compression_bounce[buf->rw]);
+		mempool_free(buf->b, &buf->c->compress.bounce[buf->rw]);
 		break;
 	}
 }
@@ -72,7 +72,7 @@ static struct bbuf __bounce_alloc(struct bch_fs *c, unsigned size, int rw)
 	if (b)
 		return (struct bbuf) { .c = c, .b = b, .type = BB_kmalloc, .rw = rw };
 
-	b = mempool_alloc(&c->compression_bounce[rw], GFP_NOFS);
+	b = mempool_alloc(&c->compress.bounce[rw], GFP_NOFS);
 	if (b)
 		return (struct bbuf) { .c = c, .b = b, .type = BB_mempool, .rw = rw };
 
@@ -95,12 +95,12 @@ static bool bio_phys_contig(struct bio *bio, struct bvec_iter start)
 	void *expected_start = NULL;
 
 	__bio_for_each_bvec(bv, bio, iter, start) {
-		if (expected_start &&
-		    expected_start != page_address(bv.bv_page) + bv.bv_offset)
+		void *bv_addr = bvec_virt(&bv);
+
+		if (expected_start && expected_start != bv_addr)
 			return false;
 
-		expected_start = page_address(bv.bv_page) +
-			bv.bv_offset + bv.bv_len;
+		expected_start = bv_addr + bv.bv_len;
 	}
 
 	return true;
@@ -109,27 +109,27 @@ static bool bio_phys_contig(struct bio *bio, struct bvec_iter start)
 static struct bbuf __bio_map_or_bounce(struct bch_fs *c, struct bio *bio,
 				       struct bvec_iter start, int rw)
 {
-	struct bio_vec bv;
-	struct bvec_iter iter;
-	unsigned nr_pages = 0;
-	struct page *stack_pages[16];
-	struct page **pages = NULL;
-	void *data;
-
 	BUG_ON(start.bi_size > c->opts.encoded_extent_max);
 
-	if (!PageHighMem(bio_iter_page(bio, start)) &&
-	    bio_phys_contig(bio, start))
+#ifndef CONFIG_HIGHMEM
+	if (bio_phys_contig(bio, start))
 		return (struct bbuf) {
 			.c	= c,
-			.b	= page_address(bio_iter_page(bio, start)) +
-				bio_iter_offset(bio, start),
+			.b	= bvec_virt(&bio_iter_iovec(bio, start)),
 			.type	= BB_none,
 			.rw	= rw
 		};
+#endif
 
+#ifdef __KERNEL__
 	/* check if we can map the pages contiguously: */
+	struct bio_vec bv;
+	struct bvec_iter iter;
+	unsigned nr_pages = 0;
+
 	__bio_for_each_segment(bv, bio, iter, start) {
+		BUG_ON(bv.bv_offset + bv.bv_len > PAGE_SIZE);
+
 		if (iter.bi_size != start.bi_size &&
 		    bv.bv_offset)
 			return bio_bounce(c, bio, start, rw);
@@ -143,7 +143,8 @@ static struct bbuf __bio_map_or_bounce(struct bch_fs *c, struct bio *bio,
 
 	BUG_ON(DIV_ROUND_UP(start.bi_size, PAGE_SIZE) > nr_pages);
 
-	pages = nr_pages > ARRAY_SIZE(stack_pages)
+	struct page *stack_pages[16];
+	struct page **pages = nr_pages > ARRAY_SIZE(stack_pages)
 		? kmalloc_array(nr_pages, sizeof(struct page *), GFP_NOFS)
 		: stack_pages;
 	if (!pages)
@@ -153,19 +154,20 @@ static struct bbuf __bio_map_or_bounce(struct bch_fs *c, struct bio *bio,
 	__bio_for_each_segment(bv, bio, iter, start)
 		pages[nr_pages++] = bv.bv_page;
 
-	data = vmap(pages, nr_pages, VM_MAP, PAGE_KERNEL);
+	void *data = vmap(pages, nr_pages, VM_MAP, PAGE_KERNEL);
 	if (pages != stack_pages)
 		kfree(pages);
 
-	if (!data)
-		return bio_bounce(c, bio, start, rw);
+	if (data)
+		return (struct bbuf) {
+			c,
+			data + bio_iter_offset(bio, start),
+			BB_vmap,
+			rw
+		};
+#endif /* __KERNEL__ */
 
-	return (struct bbuf) {
-		c,
-		data + bio_iter_offset(bio, start),
-		BB_vmap,
-		rw
-	};
+	return bio_bounce(c, bio, start, rw);
 }
 
 static struct bbuf bio_map_or_bounce(struct bch_fs *c, struct bio *bio, int rw)
@@ -186,12 +188,12 @@ static int __bio_uncompress(struct bch_fs *c, struct bio *src,
 	size_t src_len = src->bi_iter.bi_size;
 	size_t dst_len = crc.uncompressed_size << 9;
 	void *workspace;
-	int ret = 0, ret2;
+	int ret2;
 
 	enum bch_compression_opts opt = bch2_compression_type_to_opt(crc.compression_type);
-	mempool_t *workspace_pool = &c->compress_workspace[opt];
+	mempool_t *workspace_pool = &c->compress.workspace[opt];
 	if (unlikely(!mempool_initialized(workspace_pool))) {
-		if (fsck_err(c, compression_type_not_marked_in_sb,
+		if (ret_fsck_err(c, compression_type_not_marked_in_sb,
 			     "compression type %s set but not marked in superblock",
 			     __bch2_compression_types[crc.compression_type]))
 			try(bch2_check_set_has_compressed_data(c, opt));
@@ -252,8 +254,8 @@ static int __bio_uncompress(struct bch_fs *c, struct bio *src,
 	default:
 		BUG();
 	}
-fsck_err:
-	return ret;
+
+	return 0;
 }
 
 int bch2_bio_uncompress_inplace(struct bch_write_op *op,
@@ -386,7 +388,7 @@ static int attempt_compress(struct bch_fs *c,
 		 */
 		unsigned level = min((compression.level * 3) / 2, zstd_max_clevel());
 		ZSTD_parameters params = zstd_get_params(level, c->opts.encoded_extent_max);
-		ZSTD_CCtx *ctx = zstd_init_cctx(workspace, c->zstd_workspace_size);
+		ZSTD_CCtx *ctx = zstd_init_cctx(workspace, c->compress.zstd_workspace_size);
 
 		/*
 		 * ZSTD requires that when we decompress we pass in the exact
@@ -426,9 +428,9 @@ static unsigned __bio_compress(struct bch_fs *c,
 	/* bch2_compression_decode catches unknown compression types: */
 	BUG_ON(compression.type >= BCH_COMPRESSION_OPT_NR);
 
-	mempool_t *workspace_pool = &c->compress_workspace[compression.type];
+	mempool_t *workspace_pool = &c->compress.workspace[compression.type];
 	if (unlikely(!mempool_initialized(workspace_pool))) {
-		if (fsck_err(c, compression_opt_not_marked_in_sb,
+		if (ret_fsck_err(c, compression_opt_not_marked_in_sb,
 			     "compression opt %s set but not marked in superblock",
 			     bch2_compression_opts[compression.type])) {
 			ret = bch2_check_set_has_compressed_data(c, compression.type);
@@ -512,8 +514,6 @@ static unsigned __bio_compress(struct bch_fs *c,
 	BUG_ON(*dst_len & (block_bytes(c) - 1));
 	BUG_ON(*src_len & (block_bytes(c) - 1));
 	return compression_type;
-fsck_err:
-	return BCH_COMPRESSION_TYPE_none;
 }
 
 unsigned bch2_bio_compress(struct bch_fs *c,
@@ -587,10 +587,10 @@ void bch2_fs_compress_exit(struct bch_fs *c)
 {
 	unsigned i;
 
-	for (i = 0; i < ARRAY_SIZE(c->compress_workspace); i++)
-		mempool_exit(&c->compress_workspace[i]);
-	mempool_exit(&c->compression_bounce[WRITE]);
-	mempool_exit(&c->compression_bounce[READ]);
+	for (i = 0; i < ARRAY_SIZE(c->compress.workspace); i++)
+		mempool_exit(&c->compress.workspace[i]);
+	mempool_exit(&c->compress.bounce[WRITE]);
+	mempool_exit(&c->compress.bounce[READ]);
 }
 
 static int __bch2_fs_compress_init(struct bch_fs *c, u64 features)
@@ -598,7 +598,7 @@ static int __bch2_fs_compress_init(struct bch_fs *c, u64 features)
 	ZSTD_parameters params = zstd_get_params(zstd_max_clevel(),
 						 c->opts.encoded_extent_max);
 
-	c->zstd_workspace_size = zstd_cctx_workspace_bound(&params.cParams);
+	c->compress.zstd_workspace_size = zstd_cctx_workspace_bound(&params.cParams);
 
 	struct {
 		unsigned			feature;
@@ -611,7 +611,7 @@ static int __bch2_fs_compress_init(struct bch_fs *c, u64 features)
 			max(zlib_deflate_workspacesize(MAX_WBITS, DEF_MEM_LEVEL),
 			    zlib_inflate_workspacesize()) },
 		{ BCH_FEATURE_zstd, BCH_COMPRESSION_OPT_zstd,
-			max(c->zstd_workspace_size,
+			max(c->compress.zstd_workspace_size,
 			    zstd_dctx_workspace_bound()) },
 	}, *i;
 	bool have_compressed = false;
@@ -624,13 +624,13 @@ static int __bch2_fs_compress_init(struct bch_fs *c, u64 features)
 	if (!have_compressed)
 		return 0;
 
-	if (!mempool_initialized(&c->compression_bounce[READ]) &&
-	    mempool_init_kvmalloc_pool(&c->compression_bounce[READ],
+	if (!mempool_initialized(&c->compress.bounce[READ]) &&
+	    mempool_init_kvmalloc_pool(&c->compress.bounce[READ],
 				       1, c->opts.encoded_extent_max))
 		return bch_err_throw(c, ENOMEM_compression_bounce_read_init);
 
-	if (!mempool_initialized(&c->compression_bounce[WRITE]) &&
-	    mempool_init_kvmalloc_pool(&c->compression_bounce[WRITE],
+	if (!mempool_initialized(&c->compress.bounce[WRITE]) &&
+	    mempool_init_kvmalloc_pool(&c->compress.bounce[WRITE],
 				       1, c->opts.encoded_extent_max))
 		return bch_err_throw(c, ENOMEM_compression_bounce_write_init);
 
@@ -640,11 +640,11 @@ static int __bch2_fs_compress_init(struct bch_fs *c, u64 features)
 		if (!(features & (1 << i->feature)))
 			continue;
 
-		if (mempool_initialized(&c->compress_workspace[i->type]))
+		if (mempool_initialized(&c->compress.workspace[i->type]))
 			continue;
 
 		if (mempool_init_kvmalloc_pool(
-				&c->compress_workspace[i->type],
+				&c->compress.workspace[i->type],
 				1, i->compress_workspace))
 			return bch_err_throw(c, ENOMEM_compression_workspace_init);
 	}
