@@ -13,7 +13,7 @@
 
 static int bch2_snapshot_table_make_room(struct bch_fs *c, u32 id)
 {
-	guard(mutex)(&c->snapshot_table_lock);
+	guard(mutex)(&c->snapshots.table_lock);
 	return bch2_snapshot_t_mut(c, id)
 		? 0
 		: bch_err_throw(c, ENOMEM_mark_snapshot);
@@ -38,7 +38,7 @@ u32 bch2_snapshot_oldest_subvol(struct bch_fs *c, u32 snapshot_root,
 				snapshot_id_list *skip)
 {
 	guard(rcu)();
-	struct snapshot_table *t = rcu_dereference(c->snapshots);
+	struct snapshot_table *t = rcu_dereference(c->snapshots.table);
 
 	while (true) {
 		u32 id = snapshot_root, subvol = 0;
@@ -551,7 +551,7 @@ int bch2_reconstruct_snapshots(struct bch_fs *c)
 	int ret = 0;
 
 	struct progress_indicator progress;
-	bch2_progress_init(&progress, c, btree_has_snapshots_mask);
+	bch2_progress_init(&progress, __func__, c, btree_has_snapshots_mask, 0);
 
 	for (unsigned btree = 0; btree < BTREE_ID_NR; btree++) {
 		if (btree_type_has_snapshots(btree)) {
@@ -559,7 +559,7 @@ int bch2_reconstruct_snapshots(struct bch_fs *c)
 
 			try(for_each_btree_key(trans, iter, btree, POS_MIN,
 					BTREE_ITER_all_snapshots|BTREE_ITER_prefetch, k, ({
-				progress_update_iter(trans, &progress, &iter) ?:
+				bch2_progress_update_iter(trans, &progress, &iter) ?:
 				get_snapshot_trees(c, &r, k.k->p);
 			})));
 
